"""DEPRECATED: Kept for backwards compatibility."""
from langchain_core.utils.input import (
    get_bolded_text,
    get_color_mapping,
    get_colored_text,
    print_text,
)

__all__ = [
    "get_bolded_text",
    "get_color_mapping",
    "get_colored_text",
    "print_text",
]
